import java.io.*;
/**
 * AbstractBoardGame implements common methods to
 * TicTacToe and Gomoku
 * @author Oscar.Nierstrasz@acm.org
 * @version 1.6 1999-02-07
 */
public abstract class AbstractBoardGame implements BoardGame {

	protected char[][] _gameState;
	protected int _rows;
	protected int _cols;
	protected int _winningScore;

	protected Player _winner = new Player(); // nobody
	protected Player[] _player;
	protected int _turn = X; // initial turn
	protected int _squaresLeft = 9;

	static final int X = 0;
	static final int O = 1;

	/**
	 * Generic initialization for BoardGame.
	 * The state of the game is represented as 3x3
	 * array of chars marked ' ', 'X', or 'O'.
	 * We index the state using chess notation,
	 * i.e., column is 'a' through 'c' and row is
	 * '1' through '3'.
	 */
	protected void init(int rows, int cols, int score,
		Player playerX, Player playerO)
	{
		_rows = rows;
		_cols = cols;
		_gameState = new char[_rows][_cols];
		for (int col=0; col <_cols; col++)
			for (int row=0; row<_rows; row++)
				try {
					this.set(col,row,' ');
				} catch(AssertionException err) {
					System.err.println("Fatal error -- BoardGame initialization failed!");
					err.printStackTrace();
					System.exit(1);
				}
		_squaresLeft = _rows * _cols;
		
		_winningScore = score;
		
		_player = new Player[2];
		_player[X] = playerX;
		_player[O] = playerO;
	}

	/**
	 * set() and get() translate between chess coordinates
	 * and array indices.
	 */
	protected void set(int col, int row, char mark)
		throws AssertionException
	{
		assert(inRange(col, row));
		_gameState[col][row] = mark;
	}

	/**
	 * Needed by Runner
	 */
	public char get(int col, int row)
		throws AssertionException
	{
		assert(inRange(col, row));
		return _gameState[col][row];
	}

	/**
	 * Ask the current player to make a move.
	 */
	public void update() throws IOException {
		_player[_turn].move(this);
	}

	public Player currentPlayer() {
		return _player[_turn];
	}
	
	/**
	 * Called by the current player during an update().
	 * The player attempts to put a mark at a location.
	 */
	public void move(String coord, char mark)
		throws AssertionException
	{
		assert(this.notOver());
		int col = getCol(coord);
		int row = getRow(coord);
		// System.out.println("Move: <" + coord + ">");
		assert(this.get(col, row) == ' ');
		// System.out.println(mark + " at " + coord);
		this.set(col, row, mark);
		this._squaresLeft--;
		this.swapTurn();
		this.checkWinner(col,row);
		assert(this.invariant());
	}

	protected int getCol(String coord)
		throws AssertionException
	{
		assert(coord != null);
		assert(coord.length() > 0);
		char col = coord.charAt(0);
		assert((col>='a') && (col<='z'));
		return col - 'a';
	}
	
	protected int getRow(String coord)
		throws AssertionException
	{
		assert(coord != null);
		assert(coord.length() > 1);
		try {
			int row = Integer.parseInt(coord.substring(1));
			return row - 1;
		} catch (NumberFormatException err) {
			throw new AssertionException(err.getMessage());
		}
	}

	protected void swapTurn() {
		_turn = (_turn == X) ? O : X;
	}
	
	public Player winner() {
		return _winner;
	}
	
	/**
	 * The game is not over as long as there is no winner
	 * and somebody can still make a move ...
	 */
	public boolean notOver() {
		return this.winner().isNobody()
			&& this.squaresLeft() > 0;
	}
	
	public int squaresLeft() {
		return this._squaresLeft;
	}

	/**
	 * New algorithm needed for larger boards.
	 * Create a "Runner" that starts at the current position,
	 * and runs back in forth in a given direction, returning
	 * the length of the run (i.e., number of consecutive pieces
	 * of the same Player).  If the number is big enough, the
	 * current Player wins.
	 */
	protected void checkWinner(int col, int row)
		throws AssertionException
	{
		char player = this.get(col,row);
		Runner runner = new Runner(this, col, row);

		// check vertically
		if (runner.run(0,1) >= this._winningScore)
			{ this.setWinner(player); return; }
		// check horizontally
		if (runner.run(1,0) >= this._winningScore)
			{ this.setWinner(player); return; }
		// check diagonally
		if (runner.run(1,1) >= this._winningScore)
			{ this.setWinner(player); return; }
		// and check the other diagonal
		if (runner.run(1,-1) >= this._winningScore)
			{ this.setWinner(player); return; }
	}
	
	/**
	 * Look up which player is the winner, and set _winner
	 * accordingly. Hm. Maybe we should store Players
	 * instead of chars in our array!
	 */
	protected void setWinner(char player) {
		if (player == ' ')
			return;
		if (player == _player[X].mark())
			_winner = _player[X];
		else
			_winner = _player[O];
	}

	/**
	 * A plain ascii representation of the game,
	 * mainly for debugging purposes.
	 */
	public String toString() {
		StringBuffer rep = new StringBuffer();
	
		for (int row=this._rows-1; row>=0; row--) {
			rep.append(1 + row);
			if (row < 9)
				rep.append(' '); // extra space for single digit
			rep.append("  ");
			for (int col=0; col < this._cols; col++) {
				try {
					rep.append(this.get(col,row));
				} catch (AssertionException err) {
					rep.append("Assertion violation: "
						+ err.getMessage());
				}
				if (col<this._cols-1) {
					rep.append(" | ");
				}
			}
			rep.append('\n');
			if (row>0) {
				rep.append("   ---");
				for (int i=1; i<this._cols; i++)
					rep.append("+---");
				rep.append("\n");
			}
		}
		rep.append(" ");
		for (int col=0; col < this._cols; col++) {
			rep.append("   ");
			rep.append((char) ('a' + col));
		}
		rep.append("\n");
		return(rep.toString());
	}

	public boolean inRange(int col, int row) {
		return ((0<=col) && (col<this._cols)
			&& (0<=row) && (row<this._rows));
	}

	protected void assert(boolean assertion)
		throws AssertionException
	{
		if (!assertion) {
				throw new AssertionException("assertion failed in "
					+ this.getClass().getName());
		}
	}
	
	/**
	 * These seem obvious, which is exactly why
	 * they should be checked.
	 */
	private boolean invariant() {
		return (_turn == X || _turn == O)
			&& (this.notOver() 
				|| this.winner() == _player[X]
				|| this.winner() == _player[O]
				|| this.winner().isNobody())
			&& (_squaresLeft < _rows * _cols
				// else, initially:
				|| _turn == X && this.winner().isNobody());
	}

	/**
	 * Some generic tests for both TicTacToe and Gomoku.
	 */
	public void test() {
		try {
			System.err.println("Started AbstractboardGame tests");
			assert(this.invariant());

			assert(this.get(0,0) == ' ');
			this.set(0,0,'X');
			assert(this.get(0,0) == 'X');
			this.set(0,0,' ');
			assert(this.get(0,0) == ' ');

			assert(this.inRange(0,0));
			assert(this.inRange(_cols-1,_rows-1));

			assert(!this.inRange(-1,0));
			assert(!this.inRange(0,-1));
			assert(!this.inRange(0,_rows));
			assert(!this.inRange(_cols,0));

			assert(this.invariant());
			System.err.println("Passed AbstractboardGame tests");
		} catch (AssertionException err) {
			System.err.println("AbstractboardGame assertion failed!");
			err.printStackTrace();
		}
	}
}
